/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Interchange.c - interchange 2 rows in a matrix
 *
 */


#include <strings.h>
#include <string.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"

extern rowID *SP_PivotVector;		/* Pivot Vector */
extern int SP_ibufsize;

static short *_localFlag;	/* local flags for Interchange */
static ELEMTYPE *_tmpIbuffer;
static ELEMTYPE *_IinBuffer;	/* buffers which hold interchange rows from others */

void SP_InitInterchange() {
  _localFlag = (short*) SP_Malloc(SP_BlkSize*sizeof(short));
  _IinBuffer = (ELEMTYPE*) SP_Malloc(SP_BlkSize*SP_ibufsize*sizeof(ELEMTYPE));
  _tmpIbuffer = (ELEMTYPE*) SP_Malloc(SP_ibufsize*sizeof(ELEMTYPE));
}

void SP_FinalInterchange() {
  SP_Free(_localFlag);
  SP_Free(_IinBuffer);
  SP_Free(_tmpIbuffer);
}

static void _LocalInterchange(SP_SNmatrix_struct *SNma, INDEXTYPE k, INDEXTYPE i, INDEXTYPE blockNo, INDEXTYPE rowNo) {
  INDEXTYPE j, Sstep, Dstep, Sindex, Dindex;
  INDEXTYPE SblkAddr, DblkAddr, kendr;
  ELEMTYPE tmp;
  INDEXTYPE height, end;
  ELEMTYPE *tmpp;

  kendr = SNma->UstartRow[RINDEX(k)+1];

  if (k==blockNo) {
    for(SblkAddr = SNma->UstartRow[RINDEX(k)]; SblkAddr < kendr; ++SblkAddr){
      if (SNma->UnodeRow[SblkAddr]){
	Sstep = Dstep = DSIZE(SNma, k);             
	Sindex = i;
	Dindex = rowNo;
	for(j=0; j < SNma->UsubColNumInBlk[SblkAddr]; ++j){
	  tmp = SNma->UnodeRow[SblkAddr][Sindex];              
	  SNma->UnodeRow[SblkAddr][Sindex] =  SNma->UnodeRow[SblkAddr][Dindex];
	  SNma->UnodeRow[SblkAddr][Dindex] = tmp;
	  Sindex = Sindex + Sstep;
	  Dindex = Dindex + Dstep;
	}
      }
    }
    return;     
  }

  DblkAddr = SNma->LcacheRow[RINDEX(blockNo)];
  for (SblkAddr = SNma->UstartRow[RINDEX(k)]; (SblkAddr < kendr) && (SNma->UindexRow[SblkAddr] <= blockNo); ++SblkAddr) {
    while (SNma->LindexRow[DblkAddr] < SNma->UindexRow[SblkAddr])
      DblkAddr++;
    
    if (!SNma->UnodeRow[SblkAddr]) {
      if (!(tmpp=SNma->LnodeCol[SNma->Lrow2col[DblkAddr]])) continue;
      tmpp += numAhead(rowNo, SNma->LsubRow[SNma->Lrow2col[DblkAddr]]);
      height = SNma->LsubRowNumInBlk[SNma->Lrow2col[DblkAddr]];
      for(j=0, end=DSIZE(SNma, SNma->LindexRow[DblkAddr]); j<end; ++j, tmpp+=height)
	if (*tmpp < -SP_Epsilon || *tmpp > SP_Epsilon)
	  break;
      if (j==end) continue;
      SNma->UnodeRow[SblkAddr] = SP_MallocElem( SNma->UsubColNumInBlk[SblkAddr]*DSIZE(SNma, k) );     
    }
    else if (!SNma->LnodeCol[SNma->Lrow2col[DblkAddr]]) {
      tmpp = SNma->UnodeRow[SblkAddr]+i;
      height = DSIZE(SNma, k);
      for(j=0, end=SNma->UsubColNumInBlk[SblkAddr]; j<end; ++j, tmpp+=height)
	if (*tmpp < -SP_Epsilon || *tmpp > SP_Epsilon)
	  break;
      if (j==end) continue;
      SNma->LnodeCol[SNma->Lrow2col[DblkAddr]] = SP_MallocElem( SNma->LsubRowNumInBlk[SNma->Lrow2col[DblkAddr]]*DSIZE(SNma, SNma->LindexRow[DblkAddr]) );      
    }
    
    /* interchange between a U block and a L block */
    Sstep = DSIZE(SNma, k);
    Dstep = SNma->LsubRowNumInBlk[SNma->Lrow2col[DblkAddr]];
    Sindex = i;
    Dindex = numAhead(rowNo, SNma->LsubRow[SNma->Lrow2col[DblkAddr]]);
    for (j=0; j < DSIZE(SNma, SNma->UindexRow[SblkAddr]); ++j) {
      if (testcol(SNma->UsubCol[SblkAddr], j)) {
	tmp = *(SNma->UnodeRow[SblkAddr]+Sindex);
	*(SNma->UnodeRow[SblkAddr]+Sindex) = *(SNma->LnodeCol[SNma->Lrow2col[DblkAddr]]+Dindex);
	*(SNma->LnodeCol[SNma->Lrow2col[DblkAddr]]+Dindex) = tmp;
	Sindex += Sstep;
      }
      Dindex += Dstep;
    }
  }

  DblkAddr = SNma->UstartRow[RINDEX(blockNo)];
  for (;SblkAddr < kendr; ++SblkAddr) {
    while (SNma->UindexRow[DblkAddr] < SNma->UindexRow[SblkAddr]) 
	DblkAddr ++;

    if (!SNma->UnodeRow[SblkAddr]) {
      if (!SNma->UnodeRow[DblkAddr]) continue;
      tmpp = SNma->UnodeRow[DblkAddr] + rowNo;
      height = DSIZE(SNma, blockNo);
      for(j=0, end=SNma->UsubColNumInBlk[DblkAddr]; j<end; ++j, tmpp+=height)
	if (*tmpp < -SP_Epsilon || *tmpp > SP_Epsilon)
	  break;
      if (j==end) continue;
      SNma->UnodeRow[SblkAddr] = SP_MallocElem( SNma->UsubColNumInBlk[SblkAddr]*DSIZE(SNma, k) );     
    }
    else if (!SNma->UnodeRow[DblkAddr]) {
      tmpp = SNma->UnodeRow[SblkAddr]+i;
      height = DSIZE(SNma, k);
      for(j=0, end=SNma->UsubColNumInBlk[SblkAddr]; j<end; ++j, tmpp+=height)
	if (*tmpp < -SP_Epsilon || *tmpp > SP_Epsilon)
	  break;
      if (j==end) continue;
      SNma->UnodeRow[DblkAddr] = SP_MallocElem( SNma->UsubColNumInBlk[DblkAddr]*DSIZE(SNma, blockNo) );      
    }
    
    /* interchange between 2 U blocks */
    Sstep = DSIZE(SNma, k);
    Dstep = DSIZE(SNma, blockNo);
    Sindex = i;
    Dindex = rowNo;
    for (j=0; j < DSIZE(SNma, SNma->UindexRow[SblkAddr]); ++j) {
      if (testcol(SNma->UsubCol[SblkAddr], j)) {
	tmp = *(SNma->UnodeRow[SblkAddr]+Sindex);
	*(SNma->UnodeRow[SblkAddr]+Sindex) = *(SNma->UnodeRow[DblkAddr]+Dindex);
	*(SNma->UnodeRow[DblkAddr]+Dindex) = tmp;
	Sindex += Sstep;
      }
      if (testcol(SNma->UsubCol[DblkAddr], j)) Dindex += Dstep;
    }
  }
}

/* copy the row from the block to the buffer */
/* rno is the no of the row inside of the block */
static INDEXTYPE _CopyRowToBuffer(SP_SNmatrix_struct *SNma, INDEXTYPE brow, INDEXTYPE rno, ELEMTYPE *buffer) {
  INDEXTYPE size=1;
  INDEXTYPE drsize, dcsize, bcol, ecol;
  PATTYPE pattern;
  INDEXTYPE blkAddr;
  ELEMTYPE *pbuf;	/* points to the current position of the buffer */
  ELEMTYPE *pblk;	/* points to the data in the blocks to be copied */
  int zeroflag;

  /* compress all rows owned by the processor into buffer */
  pbuf = buffer;
  drsize = DSIZE(SNma, brow);

  /* copy L blocks to buffer */
  for (blkAddr = SNma->LcacheRow[RINDEX(brow)]; blkAddr < SNma->LstartRow[RINDEX(brow)+1]; blkAddr++) {
    bcol = SNma->LindexRow[blkAddr];
    *(int*)pbuf = bcol;
    pbuf++;
    pblk = SNma->LnodeCol[SNma->Lrow2col[blkAddr]];
    if (!pblk) {
      *(int*)pbuf = 1;
      pbuf++;
      size += 2;
      continue;
    }
    dcsize=DSIZE(SNma, bcol);

    /* copy a L block to buffer */
    zeroflag = 1;
    pblk += numAhead(rno,SNma->LsubRow[SNma->Lrow2col[blkAddr]]);
    for ( ecol = 1; ecol <= dcsize; ecol++) {
      pbuf[ecol] = *pblk;
      if (zeroflag && *pblk!=0) zeroflag = 0;
      pblk += SNma->LsubRowNumInBlk[SNma->Lrow2col[blkAddr]];
    }
    if (zeroflag) {
      *(int*)pbuf = 1;
      pbuf++;
      size += 2;
      continue;
    }
    
    *(int*)pbuf = 0;
    size += dcsize+2;
    pbuf += dcsize+1;
  }

  /* copy U blocks to buffer */
  for (blkAddr = SNma->UstartRow[RINDEX(brow)]; blkAddr < SNma->UstartRow[RINDEX(brow)+1]; blkAddr++) {
    bcol = SNma->UindexRow[blkAddr];
    pblk = SNma->UnodeRow[blkAddr];
    *(int*)pbuf = bcol; 
    pbuf++;
    if (!pblk) {
      *(int*)pbuf = 1;
      pbuf++;
      size += 2;
      continue;
    }
    dcsize=DSIZE(SNma, bcol);

    /* copy a U block to buffer */
    zeroflag = 1;
    pattern = SNma->UsubCol[blkAddr];
    for ( ecol = 0; ecol < dcsize;){
      if (testcol(pattern, ecol++)) {
	pbuf[ecol] = pblk[rno];
	if (zeroflag && pblk[rno]!=0) zeroflag = 0;
	pblk += drsize;
      }
      else
	pbuf[ecol] = 0.0;
    }
    if (zeroflag) {
      *(int*)pbuf = 1;
      pbuf++;
      size += 2;
      continue;
    }
    
    *(int*)pbuf = 0;
    size += dcsize+2;
    pbuf += dcsize+1;
  }

  *(int*)pbuf = SNma->snOrder+1;

  return size;
}

/* copy the row from the buffer to the block */
/* rno is the no of the row inside of the block */
static void _CopyBufferToRow(SP_SNmatrix_struct *SNma, ELEMTYPE *buffer, INDEXTYPE brow, INDEXTYPE rno) {
  INDEXTYPE drsize, dcsize, bcol, ecol, numrows; 
  PATTYPE pattern;
  INDEXTYPE blkAddr;
  ELEMTYPE *pbuf, *pblk;
  int nullflag;

  /* copy from buffer to objects*/
  pbuf = buffer;
  drsize = DSIZE(SNma, brow);

  /* copy buffer to L blocks */
  for (blkAddr = SNma->LcacheRow[RINDEX(brow)]; blkAddr < SNma->LstartRow[RINDEX(brow)+1]; blkAddr++) {
    bcol = SNma->LindexRow[blkAddr];

    /* throw a NULL test into the pbuf increment */
    while (*(int*)pbuf<bcol) pbuf += (*(int*)(pbuf+1)) ? 2 : DSIZE(SNma, *(int*)pbuf)+2;
    if (*(int*)pbuf > bcol) continue;
    nullflag = *(int*)(pbuf+1);  /* it's NULL; we'll be blasting zeros */
    pbuf+=2;
    numrows = SNma->LsubRowNumInBlk[SNma->Lrow2col[blkAddr]];

    /* if I own the block and the other block exists*/
    dcsize = DSIZE(SNma, bcol);
    if (!SNma->LnodeCol[SNma->Lrow2col[blkAddr]])
      SNma->LnodeCol[SNma->Lrow2col[blkAddr]] = SP_MallocElem( numrows * dcsize );
    pblk = SNma->LnodeCol[SNma->Lrow2col[blkAddr]] + numAhead(rno,SNma->LsubRow[SNma->Lrow2col[blkAddr]]);

    if (nullflag)
      for (ecol=0; ecol < dcsize; ++ecol, pblk += numrows)
	*pblk = 0.0;
    else
    /* copy buffer to a L block */
      for (ecol=0; ecol<dcsize; ++ecol, pblk += numrows)
	*pblk = *pbuf++;
  }

  /* copy buffer to U blocks */
  for (blkAddr = SNma->UstartRow[RINDEX(brow)]; blkAddr < SNma->UstartRow[RINDEX(brow)+1]; blkAddr++) {
    bcol = SNma->UindexRow[blkAddr];

    /* throw a NULL test into the pbuf increment */
    while (*(int*)pbuf<bcol) pbuf += (*(int*)(pbuf+1)) ? 2 : DSIZE(SNma, *(int*)pbuf)+2;
    if (*(int*)pbuf > bcol) continue;
    nullflag = *(int*)(pbuf+1);  /* it's NULL; we'll be blasting zeros */
    pbuf+=2;

    /* if I own the block and the other block exists*/
    dcsize = DSIZE(SNma, bcol);

    if (!SNma->UnodeRow[blkAddr])
      SNma->UnodeRow[blkAddr] = SP_MallocElem( SNma->UsubColNumInBlk[blkAddr] * DSIZE(SNma, brow) );

    pblk = SNma->UnodeRow[blkAddr];
    /* copy buffer to a U block */
    pattern = SNma->UsubCol[blkAddr];

    if (nullflag) {
      for ( ecol = 0; ecol < dcsize; ecol++)
	if (testcol(pattern, ecol)) {
	  pblk[rno]=0.0;
	  pblk+=drsize;
	}
    }
    else
      for ( ecol = 0; ecol < dcsize; ecol++, pbuf++)
	if (testcol(pattern, ecol)){
	  pblk[rno]=*pbuf;
	  pblk+=drsize;
	}
  }
}

/* pivoting for a supernode */
void SP_Interchange(SP_SNmatrix_struct *SNma, INDEXTYPE k) {
  MPI_Status SP_status;
  rowID *pivot;
  INDEXTYPE i, j, j2, blockNo, rowNo;
  INDEXTYPE l;
  int rpno1, rpno2, pe;
  static INDEXTYPE put_size;

  if (SP_me_cno==COWNER(k)) 
    for (l=SNma->LstartCol[CINDEX(k)]; l<SNma->LstartCol[CINDEX(k)+1]; l++)
      SNma->LcacheRow[RINDEX(SNma->LindexCol[l])] ++;

  if (SP_me_cno != COWNER(k)) SP_get_pivot_MPI(SNma, k);
  pivot = SP_PivotVector + SNma->Sstart[k];

  rpno1 = ROWNER(k);
  if (SP_me_rno == rpno1)
    bzero(_localFlag, SP_BlkSize*sizeof(short));

  /* sending for all rows in the block */
  for (i=0; i<DSIZE(SNma, k); i++) {
    blockNo = pivot[i].blockNo;
    rowNo = pivot[i].rowNo;

    if ((k == blockNo) && (i == rowNo)) continue;

    rpno2 = ROWNER(blockNo);

    /* if the two rows are both in the same processor */
    if (rpno1 == rpno2) {
      /* if I am one of the processors */
      if (rpno1 == SP_me_rno)
        _LocalInterchange(SNma, k, i, blockNo, rowNo);
    }
    else {
      if (rpno1 == SP_me_rno) { /* if I am owner */
	j=i+1;
	pe = SP_me_cno * SP_rp + rpno2;
	while (j<DSIZE(SNma, k) && (blockNo!=pivot[j].blockNo || rowNo!=pivot[j].rowNo)) j++;
	if (j==DSIZE(SNma, k)) {
	  /* copy the row into buffer */
	  put_size = _CopyRowToBuffer(SNma, k, i, _tmpIbuffer);
	  
	  if (_localFlag[i])
	    MPI_Send(_tmpIbuffer, put_size, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD );
	  else
	    MPI_Sendrecv(_tmpIbuffer, put_size, ELEMTYPE_MPI, pe, SP_MPI_TAG,
			 _IinBuffer+i*SP_ibufsize, SP_ibufsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status ); 
	}
	else {
	  _CopyRowToBuffer(SNma, k, i, _IinBuffer+j*SP_ibufsize);
	  _localFlag[j] = 1;
	  if (!_localFlag[i])
	    MPI_Recv(_IinBuffer+i*SP_ibufsize, SP_ibufsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status );
	}
	_CopyBufferToRow(SNma, _IinBuffer+i*SP_ibufsize, k, i );
      }
      else if (rpno2 == SP_me_rno) {
	j2=i-1;
	while (j2>=0 && (blockNo!=pivot[j2].blockNo || rowNo!=pivot[j2].rowNo)) j2--;
	j=i+1;
	while (j<DSIZE(SNma, k) && (blockNo!=pivot[j].blockNo || rowNo!=pivot[j].rowNo)) j++;
	if (j==DSIZE(SNma, k)) {
	  pe = SP_me_cno * SP_rp + rpno1;
	  if (j2==-1) {
	    put_size = _CopyRowToBuffer(SNma, blockNo, rowNo, _tmpIbuffer);
	    MPI_Sendrecv(_tmpIbuffer, put_size, ELEMTYPE_MPI, pe, SP_MPI_TAG,
			 _IinBuffer+i*SP_ibufsize, SP_ibufsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status );
	  }
	  else
	    MPI_Recv(_IinBuffer+i*SP_ibufsize, SP_ibufsize, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status );
	  _CopyBufferToRow(SNma, _IinBuffer+i*SP_ibufsize, blockNo, rowNo );
	}
	else if (j2==-1) {
	  pe = SP_me_cno * SP_rp + rpno1;
	  put_size = _CopyRowToBuffer(SNma, blockNo, rowNo, _tmpIbuffer);
	  MPI_Send(_tmpIbuffer, put_size, ELEMTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD );
	}
      }
    }
  } /* end for */
}
