/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * Snode.c - supernode partitioning and amalgamation
 *
 */


#include <string.h>

#include "struct.h"
#include "functions.h"
#include "globals.h"


#if 0
/* count the number of nodes in a matrix */
static INDEXTYPE _CountAll(Matrix *ma) {
  INDEXTYPE i;
       *: nndoc+gnus-help:gnus-help
  INDEXTYPE num=0;
 
  for (i=1; i<=ma->order; i++) {
    num += ma->colSize[i];
    num += ma->rowSize[i];
  }
 
  return (num - ma->order);
}
#endif
 
/* count the number of nodes in the column structure of a matrix */
static INDEXTYPE _CountCol(Matrix *ma) {
  INDEXTYPE i;
  INDEXTYPE num=0;
 
  for (i=1; i<=ma->order; i++) {
    num += ma->colSize[i];
  }
 
  return num;
}
 
/* count the number of nodes in the row structure of a matrix */
static INDEXTYPE _CountRow(Matrix *ma) {
  INDEXTYPE i;
  INDEXTYPE num=0;
 
  for (i=1; i<=ma->order; i++) {
    num += ma->rowSize[i];
  }
 
  return num;
}

#if 0
/* Supernode partitioning and amalgamation */
void SP_SnodePA_r(SP_SNmatrix_struct *SNma, Matrix *ma, Node **eforest, short rL, short rU) {
  INDEXTYPE i, num, begin;
  
  SNma->Snode = (INDEXTYPE *) SP_Malloc((SNma->eleOrder+1) * sizeof(INDEXTYPE));
  SNma->Snode[1] = 1;
  num = 1;
  begin = 1;
  for (i=2; i<=SNma->eleOrder; i++) 
    if ((i-begin < SP_BlkSize) && (eforest[i] != NULL) && (eforest[i]->index == i-1) 
	&& (ma->colSize[i] - ma->colSize[i-1] < rL)
	&& (ma->rowSize[i] - ma->rowSize[i-1] < rU))
      SNma->Snode[i] = begin-i;
    else {
      SNma->Snode[i] = ++num;
      begin = i;
    }

  SNma->Sstart = (INDEXTYPE *) SP_Malloc((num+2) * sizeof(INDEXTYPE));
  SNma->Sstart[0] = num;
  SNma->Sstart[num+1] = SNma->eleOrder+1;
  for (i=num; i>0; i--)
    if (SNma->Snode[SNma->Sstart[i+1]-1] > 0)
      SNma->Sstart[i] = SNma->Sstart[i+1]-1;
    else 
      SNma->Sstart[i] = SNma->Sstart[i+1]-1+SNma->Snode[SNma->Sstart[i+1]-1];
}
#endif

/* calculate the percentage of more nonzeros generated by supernode partitioning */
static float _Increment(Matrix *ma, INDEXTYPE begin, INDEXTYPE end) {
  INDEXTYPE actual = 0; 
  float symfac;
  INDEXTYPE i;

  for (i=begin; i<=end; i++) 
    actual += ma->colSize[i] + ma->rowSize[i] - 1;
    
  symfac = (end - begin + 1) * (end - begin - 1 + ma->colSize[end] + ma->rowSize[end]);

  return ((symfac - actual) / actual); 
}

/* Supernode partitioning and amalgamation */
void SP_SnodePA_p(SP_SNmatrix_struct *SNma, Matrix *ma, Node **eforest, float p) {
  INDEXTYPE i, num, begin;

  SNma->Snode = (INDEXTYPE *) SP_Malloc((SNma->eleOrder+1) * sizeof(INDEXTYPE));
  SNma->Snode[1] = 1;
  num = 1;
  begin = 1;
  for (i=2; i<=SNma->eleOrder; i++) {
    int shouldincrement = 1;
    if ((eforest[i] == NULL) || (eforest[i]->index != i-1)) {
      shouldincrement = 0;   /* fail due to elimination forest constraints */
    }
    if (i-begin >= SP_BlkSize) {
      shouldincrement = 0;   /* fail due to the block size limit */
    }
    if (_Increment(ma, begin, i) > p) {
      shouldincrement = 0;   /* fail due to the space inflation limit */
    }
    if (shouldincrement)
      SNma->Snode[i] = begin-i;
    else {
      SNma->Snode[i] = ++num;
      begin = i;
    }
  }

  SNma->Sstart = (INDEXTYPE *) SP_Malloc((num+2) * sizeof(INDEXTYPE));
  SNma->Sstart[0] = num;
  SNma->Sstart[num+1] = SNma->eleOrder+1;
  for (i=num; i>0; i--)
    if (SNma->Snode[SNma->Sstart[i+1]-1] > 0)
      SNma->Sstart[i] = SNma->Sstart[i+1]-1;
    else 
      SNma->Sstart[i] = SNma->Sstart[i+1]-1+SNma->Snode[SNma->Sstart[i+1]-1];
}

/* total number of blocks in L & U parts */
static INDEXTYPE _totalLblk, _totalUblk;

/* Generate supernode matrix */
Matrix *SP_SnodeMatrix(SP_SNmatrix_struct *SNma, Matrix *inMatrix) {
  Matrix *outMatrix;
  NodeList *tmp;
  INDEXTYPE i, lastcol, lastrow, begin, end;

  outMatrix = (Matrix*) SP_Malloc(sizeof(Matrix));
  outMatrix->order = SNma->snOrder = SNma->Sstart[0];
  outMatrix->col = (NodeList**) SP_Malloc((outMatrix->order+1) * sizeof(NodeList*));
  outMatrix->colSize = (INDEXTYPE*) SP_Malloc((outMatrix->order+1) * sizeof(INDEXTYPE));
  outMatrix->row = (NodeList**) SP_Malloc((outMatrix->order+1) * sizeof(NodeList*));
  outMatrix->rowSize = (INDEXTYPE*) SP_Malloc((outMatrix->order+1) * sizeof(INDEXTYPE));

  for (i=1; i<=outMatrix->order; i++) {
    outMatrix->colSize[i] = 0;
    outMatrix->col[i] = NULL;
    lastcol = SNma->Sstart[i+1]-1;
    tmp = inMatrix->col[lastcol];
    while (tmp != NULL) {
      begin = index2Snode(SNma, tmp->from);
      end = index2Snode(SNma, tmp->to);
      outMatrix->colSize[i] += SP_InsertToColumn(&(outMatrix->col[i]), begin, end);
      tmp = tmp->next;
    }
  }

  for (i=1; i<=outMatrix->order; i++) {
    outMatrix->rowSize[i] = 0;
    outMatrix->row[i] = NULL;
    lastrow = SNma->Sstart[i+1]-1;
    tmp = inMatrix->row[lastrow];
    begin = index2Snode(SNma, tmp->from)+1;
    end = index2Snode(SNma, tmp->to);
    if (begin <= end ) 
      outMatrix->rowSize[i] += SP_InsertToColumn(&(outMatrix->row[i]), begin, end);
    tmp = tmp->next;
    while (tmp != NULL) {
      begin = index2Snode(SNma, tmp->from);
      end = index2Snode(SNma, tmp->to);
      outMatrix->rowSize[i] += SP_InsertToColumn(&(outMatrix->row[i]), begin, end);
      tmp = tmp->next;
    }
  }

  _totalLblk = _CountCol(outMatrix);
  _totalUblk = _CountRow(outMatrix);

  return outMatrix;
}

/* calculate the bitmap structure of (begin, end) in nl */
static PATTYPE _BitStruct(NodeList* nl, INDEXTYPE begin, INDEXTYPE end) {
  PATTYPE retVal = 0;
  INDEXTYPE i = 0;
 
  while (i <= end-begin) {
    while ((nl != NULL) && (begin+i > nl->to))
      nl = nl->next;
    if (nl == NULL) break;
    if (begin+i >= nl->from) {
      retVal |= (1<<i);
      i++;
    }
    else
      i = nl->from - begin;
  }
 
  return retVal;
}

/* allocate space for index part 1 */
void SP_AllocIndex1(SP_SNmatrix_struct *SNma) {
  MPI_Bcast( &(SNma->eleOrder), 1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(SNma->snOrder), 1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( &_totalLblk, 1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( &_totalUblk, 1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );

  if (SP_me_no != 0) {
    SNma->Sstart = (INDEXTYPE*) SP_Malloc((SNma->snOrder+2) * sizeof(INDEXTYPE));
    SNma->Snode = (INDEXTYPE*) SP_Malloc((SNma->eleOrder+1) * sizeof(INDEXTYPE));
  }

  SNma->LstartCol = (INDEXTYPE*) SP_Malloc((SNma->snOrder+2) * sizeof(INDEXTYPE));
  SNma->LindexCol = (INDEXTYPE*) SP_Malloc(_totalLblk * sizeof(INDEXTYPE));
  SNma->LsubRow = (PATTYPE*) SP_Malloc(_totalLblk * sizeof(PATTYPE)); 
  SNma->LsubRowNumInBlk = (INDEXTYPE*) SP_Malloc(_totalLblk * sizeof(INDEXTYPE));
  SNma->LsubRowNumInCol = (INDEXTYPE*) SP_Malloc((SNma->snOrder+1) * sizeof(INDEXTYPE));

  SNma->UstartRow = (INDEXTYPE*) SP_Malloc((SNma->snOrder+2) * sizeof(INDEXTYPE));
  SNma->UindexRow = (INDEXTYPE*) SP_Malloc(_totalUblk * sizeof(INDEXTYPE));
  SNma->UsubCol = (PATTYPE*) SP_Malloc(_totalUblk * sizeof(PATTYPE));
  SNma->UsubColNumInBlk = (INDEXTYPE*) SP_Malloc(_totalUblk * sizeof(INDEXTYPE));
  SNma->UsubColNumInRow = (INDEXTYPE*) SP_Malloc((SNma->snOrder+1) * sizeof(INDEXTYPE));
}

/* broadcast data in index part 1 */
void SP_BcastIndex1(SP_SNmatrix_struct *SNma) {
  MPI_Bcast( SNma->Sstart, SNma->snOrder+2, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->Snode, SNma->eleOrder+1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );

  MPI_Bcast( SNma->LstartCol, SNma->snOrder+2, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->LindexCol, _totalLblk, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->LsubRow, _totalLblk, PATTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->LsubRowNumInBlk, _totalLblk, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->LsubRowNumInCol, SNma->snOrder+1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );

  MPI_Bcast( SNma->UstartRow, SNma->snOrder+2, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->UindexRow, _totalUblk, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->UsubCol, _totalUblk, PATTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->UsubColNumInBlk, _totalUblk, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
  MPI_Bcast( SNma->UsubColNumInRow, SNma->snOrder+1, INDEXTYPE_MPI, 0, MPI_COMM_WORLD );
}


/* generate the index part 1 */
void SP_GenIndex1(SP_SNmatrix_struct *SNma, Matrix *inMatrix, Matrix *original) {
  INDEXTYPE i, k, lastcol, lastrow, begin, end;
  INDEXTYPE j;
  NodeList *tmp;

  SNma->LstartCol[SNma->snOrder+1] = _totalLblk;
  SNma->UstartRow[SNma->snOrder+1] = _totalUblk;

  j = 0;
  for (i=1; i<=SNma->snOrder; i++) {
    SNma->LstartCol[i] = j;
    tmp = inMatrix->col[i];
    SNma->LsubRowNumInCol[i] = 0;
    while (tmp != NULL) {
      for (k=tmp->from; k<=tmp->to; k++) {
	if (i==k) {
	  SNma->LsubRow[j] = (1 << DSIZE(SNma, i)) - 1;
	  SNma->LsubRowNumInBlk[j] = DSIZE(SNma, i);
	  SNma->LsubRowNumInCol[i] += SNma->LsubRowNumInBlk[j];
	}
	else {
	  lastcol = SNma->Sstart[i+1]-1;
	  begin = SNma->Sstart[k];
	  end = SNma->Sstart[k+1]-1;
	  SNma->LsubRow[j] = _BitStruct(original->col[lastcol], begin, end);
	  SNma->LsubRowNumInBlk[j] = SP_OrderOfInt(SNma->LsubRow[j]);
          if (SNma->LsubRowNumInBlk[j] >= SP_DenseThresh * DSIZE(SNma, k)) {
            SNma->LsubRow[j] = (1 << DSIZE(SNma, k)) - 1;
            SNma->LsubRowNumInBlk[j] = DSIZE(SNma, k);
          }
	  SNma->LsubRowNumInCol[i] += SNma->LsubRowNumInBlk[j];
	}
	SNma->LindexCol[j++] = k;
      }
      tmp = tmp->next;
    }
  }

  j = 0;
  for (i=1; i<=SNma->snOrder; i++) {
    SNma->UstartRow[i] = j;
    tmp = inMatrix->row[i];
    SNma->UsubColNumInRow[i] = 0;
    while (tmp != NULL) {
      for (k=tmp->from; k<=tmp->to; k++) {
	lastrow = SNma->Sstart[i+1]-1;
	begin = SNma->Sstart[k];
	end = SNma->Sstart[k+1]-1;
	SNma->UsubCol[j] = _BitStruct(original->row[lastrow], begin, end);
	SNma->UsubColNumInBlk[j] = SP_OrderOfInt(SNma->UsubCol[j]);
        if (SNma->UsubColNumInBlk[j] >= SP_DenseThresh * DSIZE(SNma, k)) {
          SNma->UsubCol[j] = (1 << DSIZE(SNma, k)) - 1;
          SNma->UsubColNumInBlk[j] = DSIZE(SNma, k);
        }
	SNma->UsubColNumInRow[i] += SNma->UsubColNumInBlk[j];
	SNma->UindexRow[j++] = k;
      }
      tmp = tmp->next;
    }
  }
}

typedef struct block {
  INDEXTYPE index;
  INDEXTYPE blkIndex;
  INDEXTYPE next;
} Block;

/* generate row index for L */
static void _GenRowIndexInL(SP_SNmatrix_struct *SNma) {
  INDEXTYPE i, row;
  Block *space;
  INDEXTYPE top, j, *head, *tail, tmp;

  head = (INDEXTYPE*) SP_Malloc(SNma->rowNum * sizeof(INDEXTYPE));
  tail = (INDEXTYPE*) SP_Malloc(SNma->rowNum * sizeof(INDEXTYPE));
  for (i=0; i<SNma->rowNum; i++) {
    head[i] = -1;
    tail[i] = -1;
  }
  space = (Block*) SP_Malloc(SNma->LstartCol[SNma->colNum] * sizeof(Block));
  top = 0;

  SNma->LstartRow = (INDEXTYPE*) SP_Malloc((SNma->rowNum+1) * sizeof(INDEXTYPE));
  SNma->LstartRow[SNma->rowNum] = SNma->LstartCol[SNma->colNum];
  SNma->LindexRow = (INDEXTYPE*) SP_Malloc(SNma->LstartCol[SNma->colNum] * sizeof(INDEXTYPE));
  SNma->Lrow2col = (INDEXTYPE*) SP_Malloc(SNma->LstartCol[SNma->colNum] * sizeof(INDEXTYPE));  

  for (i=SP_me_cno+1; i<=SNma->snOrder; i+=SP_cp)
    for (j=SNma->LstartCol[CINDEX(i)]; j<SNma->LstartCol[CINDEX(i)+1]; j++) {
      row = RINDEX(SNma->LindexCol[j]);
      if (head[row] < 0) {
	head[row] = tail[row] = top; 
	space[top].index = i;
	space[top].blkIndex = j;
	space[top++].next = -1;
      }
      else {
	space[tail[row]].next = top;
	tail[row] = top;
	space[top].index = i;
	space[top].blkIndex = j;
	space[top++].next = -1;
      }
    }

  j = 0;
  for (i=0; i<SNma->rowNum; i++) {
    SNma->LstartRow[i] = j;
    tmp = head[i];
    while (tmp >= 0) {
      SNma->Lrow2col[j] = space[tmp].blkIndex;
      SNma->LindexRow[j++] = space[tmp].index;
      tmp = space[tmp].next;
    }
  }

  SP_Free(head);
  SP_Free(tail);
  SP_Free(space);

  SNma->LcacheRow = (INDEXTYPE*) SP_Malloc((SNma->rowNum+1) * sizeof(INDEXTYPE));
  memcpy(SNma->LcacheRow, SNma->LstartRow, (SNma->rowNum+1)*sizeof(INDEXTYPE));
}

/* generate column index for U */
static void _GenColIndexInU(SP_SNmatrix_struct *SNma) {
  INDEXTYPE i, col;
  Block *space;
  INDEXTYPE top, j, *head, *tail, tmp;

  head = (INDEXTYPE*) SP_Malloc(SNma->colNum * sizeof(INDEXTYPE));
  tail = (INDEXTYPE*) SP_Malloc(SNma->colNum * sizeof(INDEXTYPE));
  for (i=0; i<SNma->colNum; i++) {
    head[i] = -1;
    tail[i] = -1;
  }
  space = (Block*) SP_Malloc(SNma->UstartRow[SNma->rowNum] * sizeof(Block));
  top = 0;

  SNma->UstartCol = (INDEXTYPE*) SP_Malloc((SNma->colNum+1) * sizeof(INDEXTYPE));
  SNma->UstartCol[SNma->colNum] = SNma->UstartRow[SNma->rowNum];
  SNma->UindexCol = (INDEXTYPE*) SP_Malloc(SNma->UstartRow[SNma->rowNum] * sizeof(INDEXTYPE));
  SNma->Ucol2row = (INDEXTYPE*) SP_Malloc(SNma->UstartRow[SNma->rowNum] * sizeof(INDEXTYPE));  

  for (i=SP_me_rno+1; i<=SNma->snOrder; i+=SP_rp)
    for (j=SNma->UstartRow[RINDEX(i)]; j<SNma->UstartRow[RINDEX(i)+1]; j++) {
      col = CINDEX(SNma->UindexRow[j]);
      if (head[col] < 0) {
	head[col] = tail[col] = top; 
	space[top].index = i;
	space[top].blkIndex = j;
	space[top++].next = -1;
      }
      else {
	space[tail[col]].next = top;
	tail[col] = top;
	space[top].index = i;
	space[top].blkIndex = j;
	space[top++].next = -1;
      }
    }

  j = 0;
  for (i=0; i<SNma->colNum; i++) {
    SNma->UstartCol[i] = j;
    tmp = head[i];
    while (tmp >= 0) {
      SNma->Ucol2row[j] = space[tmp].blkIndex;
      SNma->UindexCol[j++] = space[tmp].index;
      tmp = space[tmp].next;
    }
  }

  SP_Free(head);
  SP_Free(tail);
  SP_Free(space);
}

/* generate the index part 2 */
void SP_GenIndex2(SP_SNmatrix_struct *SNma) {
  INDEXTYPE i;
  INDEXTYPE j, k;
  
  j=0;
  for (i=SP_me_cno+1; i<=SNma->snOrder; i+=SP_cp) {
    SNma->LstartCol[CINDEX(i)] = j;
    SNma->LsubRowNumInCol[CINDEX(i)] = 0;
    for (k=SNma->LstartCol[i]; k<SNma->LstartCol[i+1]; k++) 
      if (SP_me_rno == ROWNER(SNma->LindexCol[k])) {
	SNma->LsubRow[j] = SNma->LsubRow[k];
	SNma->LsubRowNumInBlk[j] = SNma->LsubRowNumInBlk[k];
	SNma->LsubRowNumInCol[CINDEX(i)] += SNma->LsubRowNumInBlk[j];
	SNma->LindexCol[j++] = SNma->LindexCol[k];
      }
  }
  SNma->colNum = CINDEX(i);
  SNma->LstartCol[SNma->colNum] = j;

  j=0;
  for (i=SP_me_rno+1; i<=SNma->snOrder; i+=SP_rp) {
    SNma->UstartRow[RINDEX(i)] = j;
    SNma->UsubColNumInRow[RINDEX(i)] = 0;
    for (k=SNma->UstartRow[i]; k<SNma->UstartRow[i+1]; k++) 
      if (SP_me_cno == COWNER(SNma->UindexRow[k])) {
	SNma->UsubCol[j] = SNma->UsubCol[k];
	SNma->UsubColNumInBlk[j] = SNma->UsubColNumInBlk[k];
	SNma->UsubColNumInRow[RINDEX(i)] += SNma->UsubColNumInBlk[j];
	SNma->UindexRow[j++] = SNma->UindexRow[k];
      }
  }
  SNma->rowNum = RINDEX(i);
  SNma->UstartRow[SNma->rowNum] = j;

  _GenRowIndexInL(SNma);
  _GenColIndexInU(SNma);
}

/* generate the SnodeChild array */
void SP_GenSnodeChild(SP_SNmatrix_struct *SNma, Node **eforest) {
  MPI_Status SP_status;
  INDEXTYPE i, j, child;
  Node *tmp;
  int pe;
  
  SNma->SnodeChild = (INDEXTYPE*) SP_Malloc((SNma->snOrder/SP_cp+1)*sizeof(INDEXTYPE));
  
  if (SP_me_no == 0) {
    for (i=1; i<=SNma->snOrder; i++) {
      child=0;
      for (j=0; j<DSIZE(SNma, i); j++) {
	tmp=eforest[SNma->Sstart[i]+j];
        while (tmp!=NULL && index2Snode(SNma, tmp->index)==i) tmp=tmp->next;
	if (tmp!=NULL && index2Snode(SNma, tmp->index)>child) child=index2Snode(SNma, tmp->index);
      }
      
      for (pe=COWNER(i)*SP_rp; pe<(COWNER(i)+1)*SP_rp; pe++)
	if (pe==0) 
	  SNma->SnodeChild[CINDEX(i)]=child;
	else
	  MPI_Send(&child, 1, INDEXTYPE_MPI, pe, SP_MPI_TAG, MPI_COMM_WORLD);
    } 
  }
  else
    for (i=1; i<=SNma->snOrder; i++)
      if (SP_me_no >= COWNER(i)*SP_rp && SP_me_no < (COWNER(i)+1)*SP_rp)
	MPI_Recv(SNma->SnodeChild+CINDEX(i), 1, INDEXTYPE_MPI, 0, SP_MPI_TAG, MPI_COMM_WORLD, &SP_status);
}
