/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * functions.h - internal functions declaration
 *		 external functions are declared in Splus.h
 */


#ifndef __FUNCTIONS_H_
#define __FUNCTIONS_H_

#include <stdlib.h>	/* stdlib for size_t */

#include "struct.h"

/* function declared in Utils.c */
void *SP_Malloc(size_t);
void SP_Free(void*);
void SP_FreeSize(void*, int);
int SP_TotalSpace();
ELEMTYPE *SP_MallocElem(size_t);
void SP_perror(char *file, int line, char *format, ...);
int SP_OrderOfInt(PATTYPE);
void SP_Initialize(SP_SparseMatrix*, int, int, int*, int*, SP_SNmatrix_struct**);
void SP_DestroySparseMatrix(SP_SparseMatrix*);
int SP_SparseMatrixDiagonalZero(SP_SparseMatrix*);

/* functions declaration in SnMatrix.c */
void SP_DestroySNmatrix(SP_SNmatrix_struct*);
void SP_PrintInternalMatrix(SP_SNmatrix_struct*, int, int, int, int, int);
void SP_InitData(SP_SNmatrix_struct*, SP_SparseMatrix*, int*, int*); 

/* functions declaration in SymFac.c */ 
void SP_FreeNL(NodeList*);
void SP_FreeAllNL();
void SP_FreeMatrix(Matrix*);
void SP_FreeEforest(Node**, INDEXTYPE);
INDEXTYPE SP_InsertToColumn(NodeList**, INDEXTYPE, INDEXTYPE);
Matrix *SP_SymFac(SP_SparseMatrix*, Node**, int*, int*);

/* functions declaration in Snode.c */
void SP_SnodePA_p(SP_SNmatrix_struct*, Matrix*, Node**, float);
Matrix *SP_SnodeMatrix(SP_SNmatrix_struct*, Matrix*);
void SP_AllocIndex1(SP_SNmatrix_struct*);
void SP_BcastIndex1(SP_SNmatrix_struct*);
void SP_GenIndex1(SP_SNmatrix_struct*, Matrix*, Matrix*);
void SP_GenIndex2(SP_SNmatrix_struct*);
void SP_GenSnodeChild(SP_SNmatrix_struct*, Node**);

/* functions declaration in Interchange.c */
void SP_InitInterchange();
void SP_FinalInterchange();
void SP_Interchange(SP_SNmatrix_struct*, INDEXTYPE);

/* functions declaration in Comm.c */
void SP_InitComm();
void SP_FinalComm();
void SP_send_row_MPI(SP_SNmatrix_struct*, INDEXTYPE);
INDEXTYPE SP_get_row_MPI(SP_SNmatrix_struct*, INDEXTYPE);
void SP_send_col_MPI(SP_SNmatrix_struct*, INDEXTYPE);
INDEXTYPE SP_get_col_MPI(SP_SNmatrix_struct*, INDEXTYPE);
void SP_send_pivot_MPI(SP_SNmatrix_struct*, INDEXTYPE);
void SP_get_pivot_MPI(SP_SNmatrix_struct*, INDEXTYPE);
void SP_send_linv_MPI(SP_SNmatrix_struct*, INDEXTYPE);
ELEMTYPE* SP_get_linv_MPI(SP_SNmatrix_struct*, INDEXTYPE);

/* functions declaration in Sblas.c */
void SP_InitSblas();
void SP_FinalSblas();
void SP_Sdgemmc(int, int, int, ELEMTYPE*, int, int, ELEMTYPE*, int, int, ELEMTYPE*, int, int);
void SP_Sdgemmr(int, int, int, ELEMTYPE*, int, int, ELEMTYPE*, int, int, ELEMTYPE*, int, int);

/* functions declaration in LUpFac.c */
void SP_LUP(SP_SNmatrix_struct*, rowID**);

/* functions declaration in Solve.c */
void SP_TR_Solve(SP_SNmatrix_struct*, rowID*, SP_DenseMatrix*, int*, int*);

#endif
