/********************************************************************* 
-- S+ sparse GE solver (version 1.1) --

Copyright (C) 1999 by University of California at Santa Barbara. 
Copyright (C) 2005 by University of Rochester. 
All rights reserved.

S+ team contact: Kai Shen <kshen@cs.rochester.edu>, 
                 Tao Yang <tyang@cs.ucsb.edu>  

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose, without fee, and without written agreement
is hereby granted, provided that the above copyright notice and the
following paragraph appear in all copies of this software.

This software is provided ``as is'', without any warranty, including 
any implied warranty for merchantability or fitness for a particular
purpose. Under no circumstances shall the University of California at 
Santa Barbara, the University of Rochester, or their agents be liable 
for any use of, misuse of, or inability to use this software, including 
incidental and consequential damages.
************************************************************************/

/*
 * struct.h - define the internal structures of matrices and etc.
 *	      external data structures are defined in Splus.h
 */


#ifndef __STRUCT_H_
#define __STRUCT_H_

#include <mpi.h>

#include "Splus.h"	/* external data structure */

#define PRINT(x) \
  fprintf(stdout, "proc %d: file %s: line %d:  %s\n", SP_me_no, __FILE__, __LINE__, x); \
  fflush(stdout);

#define ASSERT(x) \
  if (!(x)) { \
    fprintf(stderr, "assertion fails at file %s line %d\n", __FILE__, __LINE__); \
    exit(-1); \
  }

#define MAX(a,b) ((a)>(b)?(a):(b))

typedef int   INDEXTYPE;
typedef int   PATTYPE;

typedef char* _fcd;
#define _cptofcd(s,n) (s)

/* definition for BLAS calling interface */

/* BLAS calls with lowercase letters and a trailing underscore */
#ifdef BLAS_LCUS
#ifdef SINGLE_PRECISION
#define spSCAL(a,b,c,d) (sscal_(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (strmm_(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (sger_(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (saxpy_(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (strtri_(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (sgemm_(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (strsm_(a,b,c,d,e,f,g,h,i,j,k))
#else
#define spSCAL(a,b,c,d) (dscal_(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (dtrmm_(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (dger_(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (daxpy_(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (dtrtri_(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (dgemm_(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (dtrsm_(a,b,c,d,e,f,g,h,i,j,k))
#endif
#endif  /* end of BLAS_LCUS */

/* BLAS calls with lowercase letters */
#ifdef BLAS_LC
#ifdef SINGLE_PRECISION
#define spSCAL(a,b,c,d) (sscal(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (strmm(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (sger(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (saxpy(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (strtri(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (sgemm(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (strsm(a,b,c,d,e,f,g,h,i,j,k))
#else
#define spSCAL(a,b,c,d) (dscal(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (dtrmm(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (dger(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (daxpy(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (dtrtri(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (dgemm(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (dtrsm(a,b,c,d,e,f,g,h,i,j,k))
#endif
#endif  /* end of BLAS_LC */

/* BLAS calls with uppercase letters */
#ifdef BLAS_UC
#ifdef SINGLE_PRECISION
#define spSCAL(a,b,c,d) (SSCAL(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (STRMM(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (SGER(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (SAXPY(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (STRTRI(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (SGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (STRSM(a,b,c,d,e,f,g,h,i,j,k))
#else
#define spSCAL(a,b,c,d) (DSCAL(a,b,c,d))
#define spTRMM(a,b,c,d,e,f,g,h,i,j,k) (DTRMM(a,b,c,d,e,f,g,h,i,j,k))
#define spGER(a,b,c,d,e,f,g,h,i) (DGER(a,b,c,d,e,f,g,h,i))
#define spAXPY(a,b,c,d,e,f) (DAXPY(a,b,c,d,e,f))
#define spTRTRI(a,b,c,d,e,f) (DTRTRI(a,b,c,d,e,f))
#define spGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m) (DGEMM(a,b,c,d,e,f,g,h,i,j,k,l,m))
#define spTRSM(a,b,c,d,e,f,g,h,i,j,k) (DTRSM(a,b,c,d,e,f,g,h,i,j,k))
#endif
#endif  /* end of BLAS_UC */

/* BLAS operation FLOPS count */
#define spSCALflops(a,b,c,d) (*a)
#define spTRMMflops(a,b,c,d,e,f,g,h,i,j,k) ((*e)*(*e)*(*f))
#define spGERflops(a,b,c,d,e,f,g,h,i) (3*(*a)*(*b))
#define spAXPYflops(a,b,c,d,e,f) (2*(*a))
#define spTRTRIflops(a,b,c,d,e,f) (((*c)*(*c)*(*c)+2*(*c))/3)
#define spGEMMflops(a,b,c,d,e,f,g,h,i,j,k,l,m) (2*(*c)*(*d)*(*e))
#define spTRSMflops(a,b,c,d,e,f,g,h,i,j,k) ((*f)*(*e)*(*e))

/* MPI_Datatype definitions for Send/Recv calls */
#ifdef SINGLE_PRECISION
static const MPI_Datatype ELEMTYPE_MPI  = MPI_FLOAT;
#else
static const MPI_Datatype ELEMTYPE_MPI  = MPI_DOUBLE;
#endif
static const MPI_Datatype INDEXTYPE_MPI = MPI_INT;
static const MPI_Datatype PATTYPE_MPI   = MPI_INT;

/* tag for pt2pt communication in S+ library */
#define SP_MPI_TAG 101

#define RINDEX(x) (((x)-1)/SP_rp)
#define CINDEX(x) (((x)-1)/SP_cp)

#define ROWNER(x) (((x)-1)%SP_rp)
#define COWNER(x) (((x)-1)%SP_cp)
#define PEOWNER(x,y)  (((x)-1)%SP_rp+((y)-1)%SP_cp*SP_rp)

/* convert the index of original matrix to the index of supernode matrix */
#define index2Snode(ma, x)  (ma->Snode[x]>=0 ? ma->Snode[x] : ma->Snode[(x)+ma->Snode[x]])

#define DSIZE(ma, x)  ((ma)->Sstart[(x)+1]-(ma)->Sstart[x])

#define testcol(str, i)     ((str) & (1<<(i)))
#define testdense(str, i)   ((str) == ((1 <<(i))-1))
#define setcol(str, i)      ((str) |= (1<<(i)))

/* calculate the number of bit 1 in val before index */
#define numAhead(a, b) SP_OrderOfInt((b) & ((1 << (a)) - 1))

#ifndef _TRACKSPACE_

#define SP_Free(x) free(x)
#define SP_FreeSize(x,y) free(x)

#endif

#define SP_ERR_LOC __FILE__, __LINE__

/* structure of a single node */
typedef struct node {
  INDEXTYPE index;
  struct node *next;
} Node;

/* structure of a set of contiguous nodes */
typedef struct node_list {
  INDEXTYPE from;		/* the row number of the first entry and*/
  INDEXTYPE to;			/* the last entry of contigious nonzeros*/
  struct node_list *next;	/* next node in the same column */
} NodeList;

/* structure of the whole matrix */
typedef struct {
  INDEXTYPE order;		/* the order(column number) of the matrix */
  NodeList **col;		/* column list */
  INDEXTYPE *colSize;		/* number of nonzeros in each column */
  NodeList **row;		/* row list */
  INDEXTYPE *rowSize;		/* number of nonzeros in each row */
} Matrix;

/* data structure of supernode matrix */
typedef struct {
  INDEXTYPE eleOrder;		/* elementwise matrix order */
  INDEXTYPE snOrder;		/* supernode matrix order */
  INDEXTYPE colNum;		/* supernode column number in this processor */
  INDEXTYPE rowNum;		/* supernode row number in this processor */

  INDEXTYPE *Snode;
  INDEXTYPE *Sstart;		/* supernode index */
  INDEXTYPE *SnodeChild;	/* supernodewise elemination forest parent */

  INDEXTYPE *LstartCol;		/* column wise starting position of each column in L */
  INDEXTYPE *LindexCol;         /* column wise row index of each block in L */
  PATTYPE *LsubRow;             /* the subrow structure of each block in L */
  INDEXTYPE *LsubRowNumInBlk;   /* the number of subrows in each block in L */
  INDEXTYPE *LsubRowNumInCol;   /* the number of subrows in each column in L */
  ELEMTYPE **LnodeCol;          /* column wise starting address of each block in L */
 
  INDEXTYPE *LstartRow;         /* row wise starting position of each row in L */
  INDEXTYPE *LindexRow;         /* row wise column index of each block in L */
  INDEXTYPE *Lrow2col;          /* index exchange from row wise to column wise in L */
  INDEXTYPE *LcacheRow;         /* working space initialized as LstartRow */
 
  INDEXTYPE *UstartRow;         /* row wise starting position of each row in U */
  INDEXTYPE *UindexRow;         /* row wise column index of each block in U */
  PATTYPE *UsubCol;             /* the subcolumn structure of each block in U */
  INDEXTYPE *UsubColNumInBlk;   /* the number of subcolumns in each block in U */
  INDEXTYPE *UsubColNumInRow;   /* the number of subcolumns in each row in U */
  ELEMTYPE **UnodeRow;          /* row wise starting address of each block in U */

  INDEXTYPE *UstartCol;         /* column wise starting position of each column in U */
  INDEXTYPE *UindexCol;         /* column wise row index of each block in U */
  INDEXTYPE *Ucol2row;          /* index exchange from column wise to row wise in U */
} SP_SNmatrix_struct;

/* data structure to identify row rowNo in block blockNo */
typedef struct {
  INDEXTYPE blockNo;
  INDEXTYPE rowNo;
} rowID;

#endif
